#include <new>
#include <stdio.h>
#include <windows.h>
#include <wmsdkidl.h>
#pragma comment(lib,"wmvcore")

#ifdef _DEBUG
	void inline dbgmsg(LPCSTR m, ...)
	{
		va_list va; va_start(va,m);
		(void)vprintf(m,va);
		(void)fflush(stdout);
		va_end(va);
	}
#	define DBGCODE(x) x
#else
#	define DBGCODE(x)
#endif


#define WMTIME_UNITS_PER_MILLISEC	10000Ui64
#define WMTIME_UNITS_PER_SEC		10000000Ui64


EXTERN_C inline DWORD GetTimerBase(void)
{
	return GetTickCount();
}
EXTERN_C inline DWORD GetTimerTime(DWORD base)
{
	DWORD time = GetTickCount();
	if(time < base) {
		return time + ((DWORD)-1 - base);
	} else {
		return time - base;
	}
}

#define WAIT_START(timeout,exit) \
	DWORD _WAIT_dwTime = timeout; \
	DWORD _WAIT_dwBase = GetTimerBase(); \
	for(;;) {
#define WAIT_CHECK(timeout,exit) \
		if(_WAIT_dwTime != INFINITE) { \
			timeout = GetTimerTime(_WAIT_dwBase); \
			if(timeout < _WAIT_dwTime) timeout = _WAIT_dwTime - timeout; else goto exit; \
		} \
	}

/** @return S_FALSE: timeout; S_OK: signal */
EXTERN_C HRESULT STDAPICALLTYPE ActiveWait(DWORD dwHandles, const HANDLE* lpHandles, bool zWaitAll, DWORD dwTimeout, DWORD* pdwHandle)
{
	HRESULT hr = S_FALSE;
//	DBGCODE(dbgmsg("*** ActiveWait(%lu,$%08lX -> $%08lX,%hcALL,%ld,$%08lX)\n",dwHandles,lpHandles,dwHandles > 0 && lpHandles != NULL ? lpHandles[0] : NULL,zWaitAll?'+':'-',dwTimeout,pdwHandle));
	WAIT_START(dwTimeout,exit);
//		DBGCODE(dbgmsg("*** ActiveWait(): MsgWaitForMultipleObjects(%lu,$%08lX -> $%08lX,%hcALL,%ld)...\n",dwHandles,lpHandles,dwHandles > 0 && lpHandles != NULL ? lpHandles[0] : NULL,zWaitAll?'+':'-',dwTimeout));
		DWORD rc = MsgWaitForMultipleObjects(dwHandles,lpHandles,zWaitAll ? TRUE : FALSE,dwTimeout,QS_ALLINPUT);
//		DBGCODE(dbgmsg("*** ActiveWait(): MsgWaitForMultipleObjects(%lu,$%08lX -> $%08lX,%hcALL,%ld) >>> %lu\n",dwHandles,lpHandles,dwHandles > 0 && lpHandles != NULL ? lpHandles[0] : NULL,zWaitAll?'+':'-',dwTimeout,rc));
		if(rc == WAIT_OBJECT_0 + dwHandles) {
			for(MSG msg; PeekMessage(&msg,NULL,0,0,PM_REMOVE);) (void)DispatchMessage(&msg);
		} else if(dwTimeout != INFINITE && rc == WAIT_TIMEOUT) {
			goto exit;
		} else if(rc == WAIT_FAILED) {
			hr = HRESULT_FROM_WIN32(GetLastError()); goto exit;
		} else if(dwHandles > 0 && (
			   rc >= WAIT_OBJECT_0    && rc < WAIT_OBJECT_0    + dwHandles
			|| rc >= WAIT_ABANDONED_0 && rc < WAIT_ABANDONED_0 + dwHandles
		)) {
			if(pdwHandle) *pdwHandle = rc - (rc > WAIT_ABANDONED_0 ? WAIT_ABANDONED_0 : WAIT_OBJECT_0);
			hr = S_OK; goto exit;
		} else {
			hr = E_UNEXPECTED; goto exit;
		}
	WAIT_CHECK(dwTimeout,exit);
exit:
//	DBGCODE(dbgmsg("*** ActiveWait(%lu,$%08lX -> $%08lX,%hcALL,%ld,$%08lX -> %ld) >>> 0x%08lX\n",dwHandles,lpHandles,dwHandles > 0 && lpHandles != NULL ? lpHandles[0] : NULL,zWaitAll?'+':'-',dwTimeout,pdwHandle,S_OK == hr && pdwHandle ? *pdwHandle : -1,hr));
	return hr;
}


EXTERN_C inline HRESULT STDAPICALLTYPE InitCriticalSection(CRITICAL_SECTION& cs) throw()
{
	__try { InitializeCriticalSectionAndSpinCount(&cs,0); }
	__except(EXCEPTION_EXECUTE_HANDLER) { return GetExceptionCode(); }
	return S_OK;
}
EXTERN_C inline void STDAPICALLTYPE DeleCriticalSection(CRITICAL_SECTION& cs) throw()
{
	DeleteCriticalSection(&cs);
}
EXTERN_C inline HRESULT STDAPICALLTYPE LockCriticalSection(CRITICAL_SECTION& cs) throw()
{
	__try { EnterCriticalSection(&cs); }
	__except(EXCEPTION_EXECUTE_HANDLER) { return GetExceptionCode(); }
	return S_OK;
}
EXTERN_C inline void STDAPICALLTYPE FreeCriticalSection(CRITICAL_SECTION& cs) throw()
{
	LeaveCriticalSection(&cs);
}

class CCriticalSection : public CRITICAL_SECTION
{
	private: bool cs;
	public:  CCriticalSection(void) throw() : cs(false) {}
	public: ~CCriticalSection(void) throw() { dele(); }
	public: inline HRESULT init(void) throw() { if(cs) return S_OK; HRESULT hr = InitCriticalSection(*this); if(S_OK == hr) cs = true; return hr; }
	public: inline void    dele(void) throw() { if(cs) { DeleCriticalSection(*this); cs = false; } }
	public: inline HRESULT lock(void) throw() { return cs ? LockCriticalSection(*this) : E_FAIL; }
	public: inline void    free(void) throw() { if(cs) FreeCriticalSection(*this); }
};

class CAutoEvent
{
	private: HANDLE he;
	public:  CAutoEvent(void) throw() : he(NULL) {}
	public: ~CAutoEvent(void) throw() { dele(); }
	public: inline HRESULT init(void) throw() { if(he) return S_OK; he = CreateEvent(NULL,FALSE,FALSE,NULL); return he ? S_OK : HRESULT_FROM_WIN32(GetLastError()); }
	public: inline void    dele(void) throw() { if(he) { (void)CloseHandle(he); he = NULL; } }
#ifdef _DEBUG
	public: inline HRESULT wait(DWORD timeout) throw() { return he ? (/*dbgmsg("*** CAutoEvent::wait(): ActiveWait($%08lX)...\n",he),*/ActiveWait(1,&he,false,timeout,NULL)) : E_FAIL; }
	public: inline HRESULT sign(void) throw() { return he ? ((/*dbgmsg("*** CAutoEvent::sign(): SetEvent($%08lX)...\n",he),*/SetEvent(he)) ? S_OK : HRESULT_FROM_WIN32(GetLastError())) : E_FAIL; }
#else
	public: inline HRESULT wait(DWORD timeout) throw() { return he ? ActiveWait(1,&he,false,timeout,NULL) : E_FAIL; }
	public: inline HRESULT sign(void) throw() { return he ? (SetEvent(he) ? S_OK : HRESULT_FROM_WIN32(GetLastError())) : E_FAIL; }
#endif
};


#define DECLARE_IUNKNOWN \
	public: virtual HRESULT STDMETHODCALLTYPE QueryInterface(REFIID riid, void** ppv) throw() { return CUnknown::QueryInterface(riid,ppv); } \
	public: virtual ULONG STDMETHODCALLTYPE AddRef(void) throw() { return CUnknown::AddRef(); } \
	public: virtual ULONG STDMETHODCALLTYPE Release(void) throw() { return CUnknown::Release(); }
class CUnknown : public IUnknown
{
	private: LONG lRefCount;
	protected: LPCSTR szObjName;

	protected: CUnknown(LPCSTR szObjName)
		throw() : lRefCount(0)
	{
		this->szObjName = szObjName;
	}
	protected: virtual ~CUnknown(void) throw() {}

	protected: virtual IUnknown* GetInterface(REFIID riid) throw() { return NULL; };

	/// *** IUnknown *** ///

	public: virtual HRESULT STDMETHODCALLTYPE QueryInterface(REFIID riid, void** ppv) throw()
	{
		if(ppv == NULL) return E_POINTER; else *ppv = NULL;
		*ppv = riid == __uuidof(IUnknown) ? (IUnknown*)this : GetInterface(riid);
		return *ppv == NULL ? E_NOINTERFACE : (AddRef(),S_OK);
	}
	public: virtual ULONG STDMETHODCALLTYPE AddRef(void) throw()
	{
		return (ULONG)InterlockedIncrement(&lRefCount);
	}
	public: virtual ULONG STDMETHODCALLTYPE Release(void) throw()
	{
		LONG l = InterlockedDecrement(&lRefCount); return l > 0 ? l : ((delete this),0);
	}
};


class __declspec(uuid("{AE5A3B9B-721A-4bdb-A387-5A46EAE1EC29}")) IGenericQueue : public IUnknown
{
	public: virtual int STDMETHODCALLTYPE count(void) = 0;
	/** @param item if S_OK, item is AddRef()ed */
	public: virtual HRESULT STDMETHODCALLTYPE add(const void* item) = 0;
	/** @param pitem S_FALSE if count() < 1 */
	public: virtual HRESULT STDMETHODCALLTYPE rem(void* item) = 0;
	/** @param i <0 for last */
	public: virtual HRESULT STDMETHODCALLTYPE get(int i, void** pitem) = 0;
	/** @return ActiveWait(1,<event>,false,timeout) */
	public: virtual HRESULT STDMETHODCALLTYPE wait(DWORD timeout) = 0;
	/** SetEvent(<event>) */
	public: virtual HRESULT STDMETHODCALLTYPE signal(void) = 0;
	/** @return LockCriticalSection(<lock>) */
	public: virtual HRESULT STDMETHODCALLTYPE enter(void) = 0;
	/** FreeCriticalSection(<lock>) */
	public: virtual void STDMETHODCALLTYPE leave(void) = 0;
};

class CGenericQueue : public CUnknown, public IGenericQueue
{

	private: CCriticalSection	_lock;
	private: CAutoEvent			_signal;
	private: int				_size;
	private: int				_k, _n, _count;
	private: LPBYTE				_list;

	protected: CGenericQueue(HRESULT* phr, LPCSTR szObjName, int size) throw() : CUnknown(szObjName)
		, _size(size), _k(0), _n(0), _count(0), _list(NULL)
	{
		if(S_OK != *phr) return;

		HRESULT	hr	= S_OK;
		bool	cs	= false;
		HANDLE	he	= NULL;

		if(S_OK != (hr = _lock.init())) goto exit;
		if(S_OK != (hr = _signal.init())) goto exit;

	exit:
		*phr = hr;
	}
	protected: virtual ~CGenericQueue(void) throw()
	{
		if(_list) { for(int i = 0; i < _n; i++) FreeItem(_list + _size * ((_k + i) % _count)); delete [] _list; }
	}

	protected: virtual void STDMETHODCALLTYPE FreeItem(void* item) {}

	public: static HRESULT STDAPICALLTYPE CreateInstance(IGenericQueue** ppv, int size) throw()
	{
		HRESULT			hr	= S_OK;
		CGenericQueue*	pv	= NULL;
		if(ppv == NULL) { hr = E_POINTER; goto exit; } else *ppv = NULL;
		if(NULL == (pv = new(std::nothrow) CGenericQueue(&hr,"CGenericQueue",size))) { hr = E_OUTOFMEMORY; goto exit; } else if(S_OK != hr) goto exit;
		if(S_OK != (hr = pv->QueryInterface(__uuidof(IGenericQueue),(void**)ppv))) goto exit;
		pv = NULL;
	exit:
		if(pv) delete pv;
		return S_OK;
	}

	/// *** IUnknown *** ///

	DECLARE_IUNKNOWN
	protected: virtual IUnknown* GetInterface(REFIID riid) throw()
	{
		if(riid == __uuidof(IGenericQueue)) return (IGenericQueue*)this;
		return CUnknown::GetInterface(riid);
	}

	/// *** IGenericQueue *** ///

#ifdef _DEBUG
	private: void dbgqueue(void)
	{
		for(int i = 0; i < _count; i++) {
			bool z = false; for(int j = 0; !z && j < _n; j++) if(i == (_k + j) % _count) z = true;
			dbgmsg("\t%hc[%2d/%2d:%2d+%2d] { 0x%08lX }\n",z ? '+' : '-',i,_count,_k,_n,*(DWORD*)(_list + _size * i));
		}
	}
#endif

	public: virtual int STDMETHODCALLTYPE count(void) throw() { return _n; }
	public: virtual HRESULT STDMETHODCALLTYPE add(const void* item) throw()
	{
		HRESULT		hr		= S_OK;
		bool		cs		= false;
		if(item == NULL) { hr = E_INVALIDARG; goto exit; }
//		DBGCODE(dbgmsg("*** %hs::add({ 0x%08lX })\n",szObjName,*(DWORD*)item));
		if(S_OK != (hr = _lock.lock())) goto exit; else cs = true;
		if(_n >= _count) {
			int c = _count > 0 ? _count + _count : 10;
			LPBYTE l = new(std::nothrow) BYTE[_size * c];
			if(l == NULL) { hr = E_OUTOFMEMORY; goto exit; }
			if(_list) {
				int m = _count - _k;
				if(m < _n) {
					(void)memcpy(l + _size * 0,_list + _size * _k,_size *      m );
					(void)memcpy(l + _size * m,_list + _size *  0,_size * (_n - m));
				} else {
					(void)memcpy(l + _size * 0,_list + _size * _k,_size *  _n     );
				}
				delete [] _list;
				_k = 0;
			}
			_list = l; _count = c;
		}
		(void)memcpy(_list + _size * ((_k + _n++) % _count),item,_size);
//		DBGCODE(dbgqueue());
	exit:
//		DBGCODE(dbgmsg("*** %hs::add() >>> 0x%08lX (%hcLOCK)\n",szObjName,hr,cs?'+':'-'));
		if(cs) _lock.free();
		return hr;
	}
	public: virtual HRESULT STDMETHODCALLTYPE rem(void* item) throw()
	{
		HRESULT		hr		= S_OK;
		bool		cs		= false;
//		DBGCODE(dbgmsg("*** %hs::rem($%08lX)\n",szObjName,item));
		if(item == NULL) { hr = E_POINTER; goto exit; }
		if(S_OK != (hr = _lock.lock())) goto exit; else cs = true;
		if(_n < 1) { hr = S_FALSE; goto exit; }
		(void)memcpy(item,_list + _size * _k,_size);
		if(--_n > 0) _k = (_k + 1) % _count; else _k = 0;
//		DBGCODE(dbgmsg("*** %hs::rem() >>> { 0x%08lX }\n",szObjName,*(DWORD*)item));
//		DBGCODE(dbgqueue());
	exit:
//		DBGCODE(dbgmsg("*** %hs::rem() >>> 0x%08lX (%hcLOCK)\n",szObjName,hr,cs?'+':'-'));
		if(cs) _lock.free();
		return hr;
	}
	public: virtual HRESULT STDMETHODCALLTYPE get(int i, void** pitem) throw()
	{
		HRESULT		hr		= S_OK;
		bool		cs		= false;
//		DBGCODE(dbgmsg("*** %hs::get(%d,$%08lX)\n",szObjName,i,pitem));
		if(pitem == NULL) { hr = E_POINTER; goto exit; } else *pitem = NULL;
		if(S_OK != (hr = _lock.lock())) goto exit; else cs = true;
		if(i < 0) i = _n - 1;
		if(i < 0 || i >= _n) { hr = E_INVALIDARG; goto exit; }
		*pitem = _list + _size * ((_k + i) % _count);
//		DBGCODE(dbgmsg("*** %hs::get(%d) >>> { 0x%08lX }\n",szObjName,i,*(DWORD*)(*pitem)));
//		DBGCODE(dbgqueue());
	exit:
//		DBGCODE(dbgmsg("*** %hs::get() >>> 0x%08lX (%hcLOCK)\n",szObjName,hr,cs?'+':'-'));
		if(cs) _lock.free();
		return hr;
	}
	public: virtual HRESULT STDAPICALLTYPE wait(DWORD timeout)
	{
//		DBGCODE(dbgmsg("*** %hs::wait(%ld)\n",szObjName,timeout));
		return _signal.wait(timeout);
	}
	public: virtual HRESULT STDMETHODCALLTYPE signal(void)
	{
//		DBGCODE(dbgmsg("*** %hs::signal()\n",szObjName));
		return _signal.sign();
	}
	public: virtual HRESULT STDMETHODCALLTYPE enter(void)
	{
//		DBGCODE(dbgmsg("*** %hs::enter()\n",szObjName));
		return _lock.lock();
	}
	public: virtual void STDMETHODCALLTYPE leave(void)
	{
//		DBGCODE(dbgmsg("*** %hs::leave()\n",szObjName));
		_lock.free();
	}
};

typedef enum EReaderEvent {
	EReaderEvent_None,
	EReaderEvent_Error,		/// OnStatus(WMT_ERROR)		: hr
	EReaderEvent_Open,		/// OnStatus(WMT_OPENED)	: hr
	EReaderEvent_Start,		/// OnStatus(WMT_STARTED)	: hr
	EReaderEvent_Time,		/// OnTime()				: qwSampleTime
	EReaderEvent_Sample,	/// On[Stream]Sample()		: wStreamNumber, qwSampleTime, qwSampleDuration, dwFlags, pms
	EReaderEvent_End,		/// OnStatus(WMT_END_OF_*)	:  hr
	EReaderEvent_Stop,		/// OnStatus(WMT_STOPPED)	: hr
	EReaderEvent_Close,		/// OnStatus(WMT_CLOSED)	: hr
} EReaderEvent;
typedef struct TReaderEvent {
	EReaderEvent	cmd;
	HRESULT			hr;
	WORD			wStreamNumber;
	QWORD			qwSampleTime;
	QWORD			qwSampleDuration;
	DWORD			dwFlags;
	INSSBuffer*		pms;
} TReaderEvent;

#ifdef _DEBUG
static LPCSTR szStatusName[] = {
	"WMT_ERROR",
	"WMT_OPENED",
	"WMT_BUFFERING_START",
	"WMT_BUFFERING_STOP",
	"WMT_EOF",
	"WMT_END_OF_SEGMENT",
	"WMT_END_OF_STREAMING",
	"WMT_LOCATING",
	"WMT_CONNECTING",
	"WMT_NO_RIGHTS",
	"WMT_MISSING_CODEC",
	"WMT_STARTED",
	"WMT_STOPPED",
	"WMT_CLOSED",
};
#endif

class __declspec(uuid("{5A0FAD32-0E66-480a-8C3C-007F40F8ED33}")) IReaderCallback : public IUnknown
{
	/**
	 * @param cmd EReaderEvent_None to match any event
	 * @param pe if != NULL, FreeEvent(pe)
	 * @param pcmd if != NULL, *pcmd == pe->cmd
	 * @param phr if != NULL, *phr == pe->hr
	 * @return S_FALSE on timeout
	 */
	public: virtual HRESULT STDMETHODCALLTYPE WaitEvent(EReaderEvent cmd, DWORD timeout, TReaderEvent* pe, EReaderEvent* pcmd, HRESULT* phr) = 0;
	public: virtual void STDMETHODCALLTYPE InitEvent(TReaderEvent* pe) = 0;
	public: virtual void STDMETHODCALLTYPE FreeEvent(TReaderEvent* pe) = 0;
};

class CReaderCallback : public CUnknown, public IWMReaderCallback, public IWMReaderCallbackAdvanced, public IReaderCallback
{
	private: class CEventQueue : public CGenericQueue
	{
		friend class CReaderCallback;
		protected: CEventQueue(HRESULT* phr) throw() : CGenericQueue(phr,"CEventQueue",sizeof(TReaderEvent)) {}
		protected: virtual void STDMETHODCALLTYPE FreeItem(void* item)
		{
			FreeEventItem((TReaderEvent*)item);
		}
		public: static HRESULT STDAPICALLTYPE CreateInstance(IGenericQueue** ppv) throw()
		{
			HRESULT			hr	= S_OK;
			CEventQueue*	pv	= NULL;
			if(ppv == NULL) { hr = E_POINTER; goto exit; } else *ppv = NULL;
			if(NULL == (pv = new(std::nothrow) CEventQueue(&hr))) { hr = E_OUTOFMEMORY; goto exit; } else if(S_OK != hr) goto exit;
			if(S_OK != (hr = pv->QueryInterface(__uuidof(IGenericQueue),(void**)ppv))) goto exit;
			pv = NULL;
		exit:
			if(pv) delete pv;
			return S_OK;
		}
		private: static void STDAPICALLTYPE InitEventItem(TReaderEvent* e)
		{
			(void)memset(e,NULL,sizeof(*e));
		}
		private: static void STDAPICALLTYPE FreeEventItem(TReaderEvent* e)
		{
			if(e->pms) e->pms->Release();
			(void)memset(e,NULL,sizeof(*e));
		}
	};

	private: IGenericQueue*	pq;

	private: CReaderCallback(HRESULT* phr) throw() : CUnknown("CReaderCallback"), pq(NULL)
	{
		if(S_OK != *phr) return;

		HRESULT			hr	= S_OK;
		IGenericQueue*	pq	= NULL;

		if(S_OK != (hr = CEventQueue::CreateInstance(&pq))) goto exit;

		this->pq = pq; pq = NULL;

	exit:
		if(pq) pq->Release();
		*phr = hr;
	}
	protected: virtual ~CReaderCallback(void)
	{
		if(pq) pq->Release();
	}

	public: static HRESULT STDAPICALLTYPE CreateInstance(IWMReaderCallback** ppv) throw()
	{
		HRESULT				hr	= S_OK;
		CReaderCallback*	pv	= NULL;
		if(ppv == NULL) { hr = E_POINTER; goto exit; }
		if(NULL == (pv = new(std::nothrow) CReaderCallback(&hr))) { hr = E_OUTOFMEMORY; goto exit; } else if(S_OK != hr) goto exit;
		if(S_OK != (hr = pv->QueryInterface(__uuidof(IWMReaderCallback),(void**)ppv))) goto exit;
		pv = NULL;
	exit:
		if(pv) delete pv;
		return hr;
	}

	/// *** IUnknown *** ///

	DECLARE_IUNKNOWN
	protected: virtual IUnknown* GetInterface(REFIID riid) throw()
	{
		if(riid == __uuidof(IWMReaderCallbackAdvanced	)) return (IWMReaderCallbackAdvanced	*)this;
		if(riid == __uuidof(IWMReaderCallback			)) return (IWMReaderCallback			*)this;
		if(riid == __uuidof(IWMStatusCallback			)) return (IWMStatusCallback			*)this;
		if(riid == __uuidof(IReaderCallback				)) return (IReaderCallback				*)this;
		return CUnknown::GetInterface(riid);
	}

	/// *** IWMStatusCallback *** ///

	public: virtual HRESULT STDMETHODCALLTYPE OnStatus(WMT_STATUS status, HRESULT shr, WMT_ATTR_DATATYPE type, LPBYTE value, void* ctx) throw()
	{
		HRESULT			hr	= S_OK;
		TReaderEvent	e	= { EReaderEvent_None, shr };
		DBGCODE(dbgmsg("*** OnStatus(%2d,0x%08lX): %hs\n",status,shr,status >= 0 && status < sizeof(szStatusName)/sizeof(*szStatusName) && szStatusName[status] ? szStatusName[status] : "???"));
		switch(status)
		{
			case WMT_ERROR				: e.cmd = EReaderEvent_Error; break;
			case WMT_OPENED				: e.cmd = EReaderEvent_Open	; break;
			case WMT_STARTED			: e.cmd = EReaderEvent_Start; break;
			case WMT_END_OF_FILE		: e.cmd = EReaderEvent_End	; break; /// WMT_END_OF_FILE == WMT_EOF
			case WMT_END_OF_SEGMENT		: e.cmd = EReaderEvent_End	; break;
			case WMT_END_OF_STREAMING	: e.cmd = EReaderEvent_End	; break;
			case WMT_STOPPED			: e.cmd = EReaderEvent_Stop	; break;
			case WMT_CLOSED				: e.cmd = EReaderEvent_Close; break;
		}
		if(e.cmd != EReaderEvent_None) {
//			DBGCODE(dbgmsg("*** OnStatus(%d,0x%08lX): add(0x%08lX)\n",status,shr,e));
			if(S_OK != (hr = pq->add(&e))) goto exit;
//			DBGCODE(dbgmsg("*** OnStatus(%d,0x%08lX): signal()\n",status,shr));
			if(S_OK != (hr = pq->signal())) goto exit;
		}
	exit:
//		DBGCODE(dbgmsg("*** OnStatus(%d,0x%08lX) >>> 0x%08lX\n",status,shr,hr));
		return hr;
	}

	/// *** IWMReaderCallback *** ///

	public: virtual HRESULT STDMETHODCALLTYPE OnSample(DWORD dwOutputNum, QWORD qwSampleTime, QWORD qwSampleDuration, DWORD dwFlags, INSSBuffer* pms, void* ctx) throw()
	{
		return E_NOTIMPL;
	}

	/// *** IWMReaderCallbackAdvanced *** ///

	public: virtual HRESULT STDMETHODCALLTYPE OnStreamSample(WORD wStreamNumber, QWORD qwSampleTime, QWORD qwSampleDuration, DWORD dwFlags, INSSBuffer* pms, void* ctx) throw()
	{
		HRESULT hr = S_OK;
		TReaderEvent e = { EReaderEvent_Sample, S_OK, wStreamNumber, qwSampleTime, qwSampleDuration, dwFlags, ((pms->AddRef()),pms) };
		if(S_OK != (hr = pq->add(&e))) goto exit;
		if(S_OK != (hr = pq->signal())) goto exit;
	exit:
		return S_OK;
	}
	public: virtual HRESULT STDMETHODCALLTYPE OnTime(QWORD qwCurrentTime, void* ctx) throw()
	{
		HRESULT hr = S_OK;
		TReaderEvent e = { EReaderEvent_Time, S_OK, 0, qwCurrentTime };
		if(S_OK != (hr = pq->add(&e))) goto exit;
		if(S_OK != (hr = pq->signal())) goto exit;
	exit:
		return S_OK;
	}
	public: virtual HRESULT STDMETHODCALLTYPE OnStreamSelection(WORD wStreamCount, WORD* lpStreamNumbers, WMT_STREAM_SELECTION* lpSelections, void* ctx) throw()
	{
		return E_NOTIMPL;
	}
	public: virtual HRESULT STDMETHODCALLTYPE OnOutputPropsChanged(DWORD dwOutputNum, WM_MEDIA_TYPE* pmt, void* ctx) throw()
	{
		return E_NOTIMPL;
	}
	public: virtual HRESULT STDMETHODCALLTYPE AllocateForStream(WORD wStreamNumber, DWORD cbBuffer, INSSBuffer** ppBuffer, void* ctx) throw()
	{
		return E_NOTIMPL;
	}
	public: virtual HRESULT STDMETHODCALLTYPE AllocateForOutput(DWORD dwOutputNum, DWORD cbBuffer, INSSBuffer** ppBuffer, void* ctx) throw()
	{
		return E_NOTIMPL;
	}

	/// *** IReaderCallback *** ///

	public: virtual HRESULT STDMETHODCALLTYPE WaitEvent(EReaderEvent cmd, DWORD timeout, TReaderEvent* pe, EReaderEvent* pcmd, HRESULT* phr)
	{
		HRESULT			hr	= S_OK;
		TReaderEvent	t; InitEvent(&t);
		WAIT_START(timeout,exit);
			switch(hr = pq->rem(&t))
			{
				case S_OK:
					if(cmd == EReaderEvent_None || t.cmd == cmd) {
						if(pcmd) *pcmd = t.cmd;
						if(phr) *phr = t.hr;
						if(pe) *pe = t; else FreeEvent(&t);
						goto exit;
					} else { FreeEvent(&t); break; }
				case S_FALSE:
					if(S_OK != (hr = pq->wait(timeout))) goto exit; else break;
				default:
					goto exit;
			}
		WAIT_CHECK(timeout,exit);
	exit:
		return hr;
	}
	public: virtual void STDMETHODCALLTYPE InitEvent(TReaderEvent* pe)
	{
		CEventQueue::InitEventItem(pe);
	}
	public: virtual void STDMETHODCALLTYPE FreeEvent(TReaderEvent* pe)
	{
		CEventQueue::FreeEventItem(pe);
	}
};

typedef struct TRecWriter {
	int					iFileIndex;
	DWORD				dwStreamCount;
	QWORD				qwTimeBase;
	IWMWriter*			pRecWriter;
	IWMWriterAdvanced*	pRecAdvanced;
} TRecWriter;

class CRecWriters : public CGenericQueue
{
	protected: CRecWriters(HRESULT* phr) throw() : CGenericQueue(phr,"CRecWriters",sizeof(TRecWriter)) {}
	protected: virtual void STDMETHODCALLTYPE FreeItem(void* item)
	{
		FreeWriterItem((TRecWriter*)item);
	}
	public: static HRESULT STDAPICALLTYPE CreateInstance(IGenericQueue** ppv) throw()
	{
		HRESULT			hr	= S_OK;
		CRecWriters*	pv	= NULL;
		if(ppv == NULL) { hr = E_POINTER; goto exit; } else *ppv = NULL;
		if(NULL == (pv = new(std::nothrow) CRecWriters(&hr))) { hr = E_OUTOFMEMORY; goto exit; } else if(S_OK != hr) goto exit;
		if(S_OK != (hr = pv->QueryInterface(__uuidof(IGenericQueue),(void**)ppv))) goto exit;
		pv = NULL;
	exit:
		if(pv) delete pv;
		return S_OK;
	}
	public: static void STDAPICALLTYPE InitWriterItem(TRecWriter* w, int iFileIndex, DWORD dwStreamCount, QWORD qwTimeBase)
	{
		(void)memset(w,NULL,sizeof(*w));
		w->iFileIndex		= iFileIndex;
		w->dwStreamCount	= dwStreamCount;
		w->qwTimeBase		= qwTimeBase;
	}
	public: static void STDAPICALLTYPE FreeWriterItem(TRecWriter* w)
	{
		if(w->pRecWriter	) w->pRecWriter		->Release();
		if(w->pRecAdvanced	) w->pRecAdvanced	->Release();
		InitWriterItem(w,-1,0,0i64);
	}

	public: static HRESULT STDAPICALLTYPE AddWriterItem(IGenericQueue* q, IWMProfile* pProfile, QWORD qwTimeBase, LPCWSTR szFileBase, LPCWSTR szFileExt, int iFileIndex, int iMinDigits)
	{
		HRESULT		hr	= S_OK;
		TRecWriter	w;	InitWriterItem(&w,iFileIndex,0,qwTimeBase);
		DWORD		dwInputCount;
		WCHAR		szFile[MAX_PATH + 1];
		if(S_OK != (hr = pProfile->GetStreamCount(&w.dwStreamCount))) goto exit;
#pragma warning(push)
#pragma warning(disable: 4996)
		int len = _snwprintf(szFile,sizeof(szFile)/sizeof(*szFile) - 1,L"%ls%0*d%ls",szFileBase,iMinDigits,iFileIndex,szFileExt);
#pragma warning(pop)
		if(len < 0) { hr = HRESULT_FROM_WIN32(ERROR_INSUFFICIENT_BUFFER); goto exit; } else szFile[len] = L'\0';
		if(S_OK != (hr = WMCreateWriter(NULL,&w.pRecWriter))) goto exit;
		if(S_OK != (hr = w.pRecWriter->QueryInterface(__uuidof(IWMWriterAdvanced),(void**)&w.pRecAdvanced))) goto exit;
		if(S_OK != (hr = w.pRecWriter->SetProfile(pProfile))) goto exit;
		if(S_OK != (hr = w.pRecWriter->GetInputCount(&dwInputCount))) goto exit;
		for(DWORD dwInputIndex = 0; dwInputIndex < dwInputCount; dwInputIndex++) {
			if(S_OK != (hr = w.pRecWriter->SetInputProps(dwInputIndex,NULL))) goto exit;
		}
		if(S_OK != (hr = w.pRecWriter->SetOutputFilename(szFile))) goto exit;
		if(S_OK != (hr = w.pRecWriter->BeginWriting())) goto exit;
		if(S_OK != (hr = q->add(&w))) goto exit; else InitWriterItem(&w,-1,0,0i64);
	exit:
		FreeWriterItem(&w);
		return hr;
	}
	public: static HRESULT STDAPICALLTYPE RemWriterItem(IGenericQueue* q)
	{
		HRESULT		hr	= S_OK;
		TRecWriter	w;	InitWriterItem(&w,-1,0,0i64);
		if(S_OK != (hr = q->rem(&w))) goto exit;
		if(S_OK != (hr = w.pRecWriter->EndWriting())) goto exit;
	exit:
		FreeWriterItem(&w);
		return hr;
	}
};


#ifdef _DEBUG
void DbgWriterItem(LPCWSTR szArgTargetBase, LPCWSTR szArgTargetExt, int iTargetDigits, int iRecIndex)
{
	int iTargetLen = wcslen(szArgTargetBase) + wcslen(szArgTargetExt) + iTargetDigits;
	dbgmsg("*** >>> %.*hs <<<\n",iTargetLen,"=========================================================================================================================");
	dbgmsg("*** >>> %ls%0*d%ls <<<\n",szArgTargetBase,iTargetDigits,iRecIndex,szArgTargetExt);
	dbgmsg("*** >>> %.*hs <<<\n",iTargetLen,"=========================================================================================================================");
}
#endif


enum { ARGPROG, ARG_SLIP_SECS, ARG_SOURCE, ARG_TARGET_BASE, ARG_TARGET_EXT, ARG_TARGET_FIRST, ARG_TARGET_DIGITS, ARGCOUNT };


int wmain(int argc, LPCWSTR argv[]) throw()
{
	HRESULT				hr				= S_OK;
	bool				com				= false;
	IWMReader*			pSrcReader		= NULL;
	IWMReaderAdvanced*	pSrcAdvanced	= NULL;
	IWMProfile*			pSrcProfile		= NULL;
	IWMStreamConfig*	pSrcStream		= NULL;
	IWMReaderCallback*	pSrcCallback	= NULL;
	IReaderCallback*	pSrcContext		= NULL;
	typedef struct TRecStream {
		int iRecIndex;
	} TRecStream;
	TRecStream*			pRecStreams		= NULL;
	IGenericQueue*		pRecWriters		= NULL;
	int					iRecFirst		= 0;
	int					iRecIndex		= 0;
	DWORD				dwStreamCount;
	WORD				wStreamNumber;
	WORD				wStreamNumbers	= 0;
	WORD				wFirstAudio		= 0;
	WORD				wFirstVideo		= 0;
	WORD				wMasterStream	= 0;
	TReaderEvent		e;
	QWORD				qwStreamTime	= 0i64;
	QWORD				qwSlipTime		= -1i64;
	LPCWSTR				szArgSource;
	LPCWSTR				szArgTargetBase;
	LPCWSTR				szArgTargetExt;
	int					iTargetFirst;
	int					iTargetDigits;

//	DBGCODE(if(!IsDebuggerPresent()) DebugBreak(););

	if(argc != ARGCOUNT) { hr = E_INVALIDARG; goto exit; }

	szArgSource		= argv[ARG_SOURCE		];
	szArgTargetBase	= argv[ARG_TARGET_BASE	];
	szArgTargetExt	= argv[ARG_TARGET_EXT	];
	iTargetFirst	= _wtoi(argv[ARG_TARGET_FIRST	]); if(iTargetFirst		< 0) { hr = E_INVALIDARG; goto exit; }
	iTargetDigits	= _wtoi(argv[ARG_TARGET_DIGITS	]); if(iTargetDigits	< 1) { hr = E_INVALIDARG; goto exit; }
	qwSlipTime		= _wtoi(argv[ARG_SLIP_SECS		]); if(*(LONGLONG*)&qwSlipTime < 0) { hr = E_INVALIDARG; goto exit; } else qwSlipTime *= WMTIME_UNITS_PER_SEC;

	DBGCODE(dbgmsg("*** %I64u secs, \"%ls\" >>> \"%ls\" %0*d \"%ls\"\n",qwSlipTime / WMTIME_UNITS_PER_SEC,szArgSource,szArgTargetBase,iTargetDigits,iTargetFirst,szArgTargetExt));

//	DBGCODE(dbgmsg("*** CoInitializeEx()...\n"));
	if(FAILED(hr = CoInitializeEx(NULL,COINIT_MULTITHREADED))) goto exit; else com = true;

//	DBGCODE(dbgmsg("*** WMCreateReader()...\n"));
	if(S_OK != (hr = WMCreateReader(NULL,WMT_RIGHT_PLAYBACK,&pSrcReader))) goto exit;
	if(S_OK != (hr = pSrcReader->QueryInterface(__uuidof(IWMReaderAdvanced),(void**)&pSrcAdvanced))) goto exit;

//	DBGCODE(dbgmsg("*** CReaderCallback::CreateInstance()...\n"));
	if(S_OK != (hr = CReaderCallback::CreateInstance(&pSrcCallback))) goto exit;
	if(S_OK != (hr = pSrcCallback->QueryInterface(__uuidof(IReaderCallback),(void**)&pSrcContext))) goto exit;
	pSrcContext->InitEvent(&e);

//	DBGCODE(dbgmsg("*** IWMReader::Open(\"%ls\")...\n",szArgSource));
	if(S_OK != (hr = pSrcReader->Open(szArgSource,pSrcCallback,NULL))) goto exit;
//	DBGCODE(dbgmsg("*** CReaderCallback::WaitEvent(Open)...\n"));
	if(S_OK != (hr = pSrcContext->WaitEvent(EReaderEvent_Open,INFINITE,NULL,NULL,&e.hr))) goto exit; else if(S_OK != e.hr) { hr = e.hr; goto exit; }

//	DBGCODE(dbgmsg("*** IWMReader::QueryInterface(IWMProfile)...\n"));
	if(S_OK != (hr = pSrcReader->QueryInterface(__uuidof(IWMProfile),(void**)&pSrcProfile))) goto exit;
	if(S_OK != (hr = pSrcProfile->GetStreamCount(&dwStreamCount))) goto exit;
//	DBGCODE(dbgmsg("*** IWMProfile::GetStreamCount() >>> %lu\n",dwStreamCount));
	for(DWORD dwStreamIndex = 0; dwStreamIndex < dwStreamCount; dwStreamIndex++) {
		if(S_OK != (hr = pSrcProfile->GetStream(dwStreamIndex,&pSrcStream))) goto exit;
		if(S_OK != (hr = pSrcStream->GetStreamNumber(&wStreamNumber))) goto exit;
		if(S_OK != (hr = pSrcAdvanced->SetReceiveStreamSamples(wStreamNumber,TRUE))) goto exit;
		GUID gStreamType;
		if(S_OK != (hr = pSrcStream->GetStreamType(&gStreamType))) goto exit;
		; ;;;; if(gStreamType == WMMEDIATYPE_Video) {
			if(wFirstVideo == 0) wFirstVideo = wStreamNumber;
		} else if(gStreamType == WMMEDIATYPE_Audio) {
			if(wFirstAudio == 0) wFirstAudio = wStreamNumber;
		}
		pSrcStream->Release(); pSrcStream = NULL;
		if(wStreamNumber > wStreamNumbers) wStreamNumbers = wStreamNumber;
	}
	if(S_OK != (hr = pSrcAdvanced->SetUserProvidedClock(TRUE))) goto exit;
	wMasterStream = wFirstVideo > 0 ? wFirstVideo : wFirstAudio > 0 ? wFirstAudio : 0;

	if(NULL == (pRecStreams = new(std::nothrow) TRecStream[++wStreamNumbers])) { hr = E_OUTOFMEMORY; goto exit; }
	for(wStreamNumber = 0; wStreamNumber < wStreamNumbers; wStreamNumber++) {
		pRecStreams[wStreamNumber].iRecIndex = 0;
	}

	if(S_OK != (hr = CRecWriters::CreateInstance(&pRecWriters))) goto exit;
	DBGCODE(DbgWriterItem(szArgTargetBase,szArgTargetExt,iTargetDigits,iTargetFirst + iRecIndex));
	if(S_OK != (hr = CRecWriters::AddWriterItem(pRecWriters,pSrcProfile,0Ui64,szArgTargetBase,szArgTargetExt,iTargetFirst + iRecIndex++,iTargetDigits))) goto exit;

	if(S_OK != (hr = pSrcReader->Start(0,0,1.0f,NULL))) goto exit;
	if(S_OK != (hr = pSrcContext->WaitEvent(EReaderEvent_Start,INFINITE,NULL,NULL,&e.hr))) goto exit; else if(S_OK != e.hr) { hr = e.hr; goto exit; }

	if(S_OK != (hr = pSrcAdvanced->DeliverTime(qwStreamTime += WMTIME_UNITS_PER_SEC))) goto exit;

	for(bool ok = true; ok;) {
		if(S_OK != (hr = pSrcContext->WaitEvent(EReaderEvent_None,INFINITE,&e,NULL,NULL))) goto exit;
		switch(e.cmd)
		{
			case EReaderEvent_Sample: {
				TRecStream* s = &pRecStreams[e.wStreamNumber];
				TRecWriter* w = NULL;
				int i = s->iRecIndex - iRecFirst;
				if(S_OK != (hr = pRecWriters->get(i,(void**)&w))) goto exit;
				QWORD t = e.qwSampleTime - w->qwTimeBase;
				if(t >= qwSlipTime && (e.dwFlags & (WM_SF_CLEANPOINT|WM_SF_DISCONTINUITY))) {
					if(wMasterStream == 0 || wMasterStream == e.wStreamNumber || s->iRecIndex + 1 < iRecIndex) {
						if(w->dwStreamCount > 1) {
							w->dwStreamCount--;
						} else {
							w->dwStreamCount = 0;
							if(i == 0) {
								if(S_OK != (hr = CRecWriters::RemWriterItem(pRecWriters))) goto exit;
								iRecFirst++;
							}
						}
						if(++s->iRecIndex >= iRecIndex) {
							DBGCODE(DbgWriterItem(szArgTargetBase,szArgTargetExt,iTargetDigits,iTargetFirst + iRecIndex));
							if(S_OK != (hr = CRecWriters::AddWriterItem(pRecWriters,pSrcProfile,e.qwSampleTime,szArgTargetBase,szArgTargetExt,iTargetFirst + iRecIndex++,iTargetDigits))) goto exit;
						}
						i = s->iRecIndex - iRecFirst;
						if(S_OK != (hr = pRecWriters->get(i,(void**)&w))) goto exit;
						t = e.qwSampleTime - w->qwTimeBase;
						e.dwFlags &= WM_SF_DISCONTINUITY;
					}
				}
				DBGCODE(dbgmsg("*** [%2hu] %hs %hs %hs %7I64d.%04I64d + %3I64d.%04I64d   >>>   %0*d: %7I64d.%04I64d\n",e.wStreamNumber
					,e.dwFlags & WM_SF_CLEANPOINT		? "SYNC" : "...."
					,e.dwFlags & WM_SF_DISCONTINUITY	? "DISC" : "...."
					,e.dwFlags & WM_SF_DATALOSS			? "LOSS" : "...."
					,e.qwSampleTime/WMTIME_UNITS_PER_MILLISEC,e.qwSampleTime%WMTIME_UNITS_PER_MILLISEC
					,e.qwSampleDuration/WMTIME_UNITS_PER_MILLISEC,e.qwSampleDuration%WMTIME_UNITS_PER_MILLISEC
					,iTargetDigits,w->iFileIndex
					,t/WMTIME_UNITS_PER_MILLISEC,t%WMTIME_UNITS_PER_MILLISEC
				));
				if(S_OK != (hr = w->pRecAdvanced->WriteStreamSample(e.wStreamNumber,t,0,e.qwSampleDuration,e.dwFlags,e.pms))) goto exit;
			} break;
			case EReaderEvent_Time:
				if(S_OK != (hr = pSrcAdvanced->DeliverTime(qwStreamTime += WMTIME_UNITS_PER_SEC))) goto exit;
				break;
			case EReaderEvent_End: ok = false; break;
			default: if(S_OK != e.hr) { hr = e.hr; goto exit; }
		}
		pSrcContext->FreeEvent(&e);
	}

	while(S_OK == (hr = CRecWriters::RemWriterItem(pRecWriters))); if(S_FALSE != hr) goto exit; else hr = S_OK;

	if(S_OK != (hr = pSrcReader->Stop())) goto exit;
	if(S_OK != (hr = pSrcContext->WaitEvent(EReaderEvent_Stop,INFINITE,NULL,NULL,&e.hr))) goto exit; else if(S_OK != e.hr) { hr = e.hr; goto exit; }

	if(S_OK != (hr = pSrcReader->Close())) goto exit;
	if(S_OK != (hr = pSrcContext->WaitEvent(EReaderEvent_Close,INFINITE,NULL,NULL,&e.hr))) goto exit; else if(S_OK != e.hr) { hr = e.hr; goto exit; }

exit:
	if(pRecStreams) delete [] pRecStreams;
	if(pRecWriters) pRecWriters->Release();
	if(pSrcContext) {
		pSrcContext->FreeEvent(&e);
		pSrcContext->Release();
	}
	if(pSrcCallback	) pSrcCallback	->Release();
	if(pSrcStream	) pSrcStream	->Release();
	if(pSrcProfile	) pSrcProfile	->Release();
	if(pSrcAdvanced	) pSrcAdvanced	->Release();
	if(pSrcReader	) pSrcReader	->Release();
	if(com) CoUninitialize();
	return hr;
}
